/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#endif
#include "debug.h"
#include "main.h"
#include "cpu.h"
#include "mem.h"
#include "screen.h"
#include "version.h"
#ifdef USE_SVGALIB
#include "svga.h"
#endif
#include "audio.h"
#if MIDI
#include "midi.h"
#endif
#include "native.h"
#include "gemdos.h"
#include "fdc.h"
#include "sfp.h"
#include "ikbd.h"
#include "ui.h"
#include "utils.h"
#include "io.h"

#define STONXRC ".stonxrc"
#define EQ(_x,_y) (is_rc ? (strcmp(_x,_y)==0) : (strcmp(_x,"-"_y)==0) )

int gr_mode=MODE_MONO;
int swidth=640, sheight=400;
extern int scr_width, scr_height, scr_planes;
int custom_screensize=0;
int boot_dev=0;
int vblpt=4;	/* interrupts per VBL */
int refresh=2;	/* VBLs per refresh */
int hz200pt=1;	/* interrupts per timer C interrupt */
long tv_usec=10000;
int verbose=1;
int vdi=0;
int vdi_mode=0;
int audio=1;
int midi=1;
int scanlines=200;
int warmboot=0;
int priv_cmap=0;
int redirect_cconws=0;
int fix_screen=0;
int realtime=0;
extern char *serial_dev;
extern char *parallel_dev;
extern int shmflag;
int got_drive=0;
int chunky=0;
extern char *kdefsfile;
Machine_Specific machine;
int timer_a=0;
int no_cart=0;

void process_args (int argc, char *argv[], int is_rc)
{
	int i, u=0;
	for (i=1; i<argc; i++)
	{
		char *x = argv[i];
		if (EQ(x,"mono")) gr_mode = MODE_MONO;
		else if (EQ(x,"color")) gr_mode = MODE_COLOR;
		else if (EQ(x,"colour")) gr_mode = MODE_COLOR;
		else if (EQ(x,"size"))
		{
			if (i == argc-1)
				error("size needs an argument of the form <width>x<height>\n");
			sscanf(argv[++i],"%dx%d",&swidth,&sheight);
			if (swidth <= 0 || sheight <= 0)
				error("invalid argument to -size\n");
			custom_screensize=1;
		}
		else if (EQ(x,"vbl"))
		{
			int v;
			if (i == argc-1 || (v = atoi(argv[++i])) <= 0)
				error("vbl needs a numeric argument > 0\n");
			else vblpt=v;
		}
		else if (EQ(x, "q"))
		{
			verbose=0;
		}
		else if (EQ(x,"private"))
		{
			priv_cmap=1;
		}
#ifdef	USE_SVGALIB
		else if (EQ(x,"svga"))
		{
			machine.screen_open = (void (*)(void))svga_screen_open;
			machine.screen_close = svga_screen_close;
			machine.screen_shifter = svga_screen_shifter;
			machine.init_keys = svga_init_keys;
			
		}
#endif
		else if (EQ(x,"chunky"))
		{
			chunky=1;
		}
		else if (EQ(x,"kmap"))
		{
			if (i == argc-1)
				error("kmap needs a file name as argument (default: Keysyms)\n");
			kdefsfile = strdup(argv[++i]);
		}
		else if (EQ(x,"timer-a"))
		{
			timer_a=1;
		}
		else if (EQ(x,"nocart"))
		{
			no_cart=1;
		}
		else if (EQ(x, "refresh"))
		{
			int t;
			if (i == argc-1 || (t = atoi(argv[++i])) <= 0)
				error("timer-c needs a numeric argument > 0\n");
			else refresh=t;
		}
		else if (EQ(x,"timer-c"))
		{
			int t;
			if (i == argc-1 || (t = atoi(argv[++i])) <= 0)
				error("timer-c needs a numeric argument > 0\n");
			else hz200pt=t;
		}
		else if (EQ(x,"usec"))
		{
			long x;
			if (i == argc-1 || (x = atol(argv[++i])) <= 0)
				error("usec needs a numeric argument > 0\n");
			else tv_usec = x;
		}
		else if (EQ(x,"vdi"))
		{
			error("-vdi has been disabled for this version (sorry!)");
			vdi_mode = vdi = 1;
		}
		else if (EQ(x,"scanlines"))
		{
			int x;
			if (i == argc-1 || (x = atoi(argv[++i])) <= 0)
				error("scanlines needs a numeric argument > 0\n");
			else scanlines = x;
		}
		else if (EQ(x,"boot"))
		{
			if (i == argc-1 || !isalpha(argv[++i][0]))
				error("boot needs a drive letter as argument (A for example)\n");
			else boot_dev=toupper(argv[i][0])-'A';
		}
		else if (EQ(x,"disk"))
		{
			char d;
			char fname[256];	/* use value from limits.h? */
			if (i == argc-1 || sscanf(argv[++i],"%c:%s", &d, fname) < 2)
				error("disk needs an argument of the form <X>:<File> where"
					" <X> is a drive letter\n");
			if (!add_drive (d, fname)) u=1;
			else got_drive=1;
		}
		else if (EQ(x,"fs"))
		{
			char d;
			char fname[256];	
			if (i == argc-1 || sscanf(argv[++i],"%c:%s", &d, fname) < 2)
				error("fs needs an argument of the form <X>:<directory> where"
					" <X> is a drive letter\n");
			if (!add_gemdos_drive (d, fname)) u=1;
			else got_drive=1;
		}
		else if (EQ(x,"serial"))
		{
			if (i == argc-1)
				error("serial needs a device as argument (/dev/ttyS1 for example)\n");
			serial_dev = strdup(argv[++i]);
		}
		else if (EQ(x,"para"))
		{
			if (i == argc-1)
				error("para needs a device as argument (/dev/lp0 for example)\n");
			parallel_dev = strdup(argv[++i]);
		}
		else if (EQ(x,"noaudio"))
		{
			audio=0;
		}
		else if (EQ(x,"nomidi"))
		{
			midi=0;
		}
		else if (EQ(x,"noshm"))
		{
			shmflag=0;
		}
		else if (EQ(x,"realtime"))
		{
			realtime=1;
		}
		else if (EQ(x,"norealtime"))
		{
			realtime=0;
		}
		else if (EQ(x,"warmboot"))
		{
			warmboot=1;
		}
		else if (EQ(x,"cconws"))
		{
			redirect_cconws=1;
		}
		else if (EQ(x,"h") || EQ(x,"?") || EQ(x,"-help"))
			u=1;
		else
		{
			fprintf(stderr, "unknown option %s\n", argv[i]);
			u=1;
		}

	}
	if (u)
	{
		if (is_rc)
			fprintf(stderr,"Error in "STONXRC", please fix!\n");
		else
		(void)fprintf (stderr, "usage: %s [<options>]\n"
			"<options> = \n"
" -q                        Be Quiet\n"
" -size <W>x<H>             (NYI)\n"
" -disk <X>:<File>          Load diskfile <File> as Drive <X>:\n"
" -fs <X>:<Directory>       Map <Directory> to Drive <X>:\n"
" -colo[u]r|-mono           Set `monitor type'\n"
" -usec <X>                 Set <X> microsecond interval timer\n"
" -vbl <X>                  Set <X> * usec VBL interval\n"
#if DOUBLE_HZ200
" -timer-c <X>              Set <X> * usec / 2 Timer C interval\n"
#else
" -timer-c <X>              Set <X> * usec Timer C interval\n"
#endif
" -refresh <X>              Set <X> * VBL refresh interval\n"
" -para <file>              Use <file> as the parallel port device\n"
" -serial <file>            Use <file> as the serial port device\n"
" -kmap   <file>            Load the Keyboard mappings from <file>\n"
" -noaudio                  Disable Audio driver\n"
" -noshm                    Disable XShm extension\n"
" -chunky                   Use `chunky' update mode\n"
" -warmboot                 Attempt to simulate a warm-boot\n"
" -private                  Use private colormap (speeds -color up)\n"
" -cconws                   Redirect GEMDOS Cconws() function to terminal\n"
#ifdef	USE_SVGALIB
" -svga                     Use Linux SVGALIB instead of X\n"	
#endif
" -vdi                      Use xlib-VDI\n\n"
"Example: %s -disk a:Disk -color -usec 10000 -vbl 10 -refresh 2 -timer-c 1\n"
"will use `Disk' for drive A:, boot in color mode, and run Timer C at "
#if DOUBLE_HZ200
"200Hz"
#else
"100Hz"
#endif
",\nVBLs at "
"10Hz"
", and refresh with "
"5Hz"
", if your system allows interval timers with\n10 ms timer intervals (see setitimer(2))\n\n", argv[0], argv[0]);
		exit(0);
	}
}

void process_stonxrc(void)
{	char *home;
    	if ((home=getenv("HOME")) != NULL)
	{	FILE *rc;	
		char stonxrc[256];
		strcpy (stonxrc,home);
		strcat (stonxrc,"/"STONXRC);
		if ((rc=fopen(stonxrc,"r")) != NULL)
		{	int i,c=1;	
			char *args[64];
			for (i=0;i<64;i++) args[i] = (char*)malloc(sizeof(char)*256);
			while (fscanf(rc,"%s",args[c]) != EOF) {
				if (args[c][0] == '#') {
					int ch;
					while ((ch = fgetc(rc)) != EOF &&
						ch != '\n')
						;
					continue;
				}
				c++;
			}
			fclose(rc);
			fprintf (stderr, "Using configuration file:%s\n", stonxrc);
			process_args (c, args,1);
			for (i=1 ;i<64;i++) free(args[i]);
		}
	}
}

void print_timing(void)
{
#ifdef DOUBLE_HZ200
#define MVAL0 2000000.0
#else
#define MVAL0 1000000.0
#endif
	if (verbose) 
	{	fprintf (stderr, "%ld microseconds interval timer\n", tv_usec);
		fprintf (stderr, "%.0f Hz Timer C, ", (double) MVAL0/(tv_usec*hz200pt));
		fprintf (stderr, "%.0f Hz VBLs and %.0f Hz screen update\n",
				(double)1000000.0/(vblpt*tv_usec),
				(double)1000000.0/(vblpt*refresh*tv_usec));
	}
}


int main (int argc, char *argv[])
{
#if defined(__NeXT__)
	fprintf(stderr,"NOTE: Co-Xist apparently has problems with virtual timers,"
				   " so I set the\n-realtime flag. Use the option -norealtime"
				   " if this bothers you.\n");
	realtime=1;
#endif
	flags = 0;

	machine.screen_open = x_screen_open;
	machine.screen_close = x_screen_close;
	machine.screen_shifter = x_screen_shifter;
	machine.init_keys = x_init_keys;

	printf ("--------------------------------------------------------------------------------\nSTonX Version " VERSION "\n(c)1994-1997 by Marinos Yannikos and Martin D. Griffiths\n"
	"This program is free software, and comes with NO WARRANTY!\n"
	"Read the file COPYING for details. If this copy of STonX was not accompanied\nby a file called `COPYING', containing the GNU GPL text, report this to the\nauthors at nino@complang.tuwien.ac.at immediately!\n--------------------------------------------------------------------------------");
	fflush(stdout);
	process_stonxrc ();
	process_args (argc, argv, 0);
	if (gr_mode == MODE_COLOR && vdi)
	{
		fprintf(stderr,"Warning: -colour -vdi implies -private!\n");
		priv_cmap=1;
	}
	if (!custom_screensize && gr_mode == MODE_COLOR)
	{
		swidth=320;
		sheight=200;
		fix_screen=0;
	}
	if ((gr_mode == MODE_MONO && (swidth != 640 || sheight != 400)) 
		|| (gr_mode == MODE_COLOR && (swidth != 320 || sheight != 200)))
		fix_screen=1;
	if (vdi && custom_screensize)
	{
		int i;
		for (i=0; i<3; i++)
		{
			scr_def[i].w = swidth;
			scr_def[i].h = sheight;
		}
	}
	else if (gr_mode == MODE_COLOR)
	{
		scr_def[0].w = swidth;
		scr_def[1].w = scr_def[2].w = swidth*2;
		scr_def[0].h = scr_def[1].h = sheight;
		scr_def[2].h = sheight*2;
	}
	else /* mode == MONO or no custom screen size or no VDI */
	{
		scr_def[0].w = swidth/2;
		scr_def[1].w = scr_def[2].w = swidth;
		scr_def[0].h = scr_def[1].h = sheight/2;
		scr_def[2].h = sheight;
	}
	/* need these for init_mem() */
	if (gr_mode == MODE_COLOR)
	{
		scr_width=scr_def[0].w;
		scr_height=scr_def[0].h;
		scr_planes=4;
	}
	else
	{
		scr_width=scr_def[2].w;
		scr_height=scr_def[2].h;
		scr_planes=1;
	}
	print_timing();
#ifndef _WIN32
	if (!got_drive)
	{
		char path[500];
		getcwd(path,500);
		(void)fprintf (stderr, "No drives specified - using default: `-fs C:%s'\n", path);
		add_gemdos_drive ('C',path);
	}
#endif
	init_debug();
	init_ui();
	init_fdc();
	init_sfp();
	init_mem();
	load_file ("tos.img",(char *)MEM(TOSSTART));
#if CARTRIDGE 
	if (!no_cart)
	{
		load_file ("cartridge.img", (char *)MEM(0xfa0000));
	}
/*	SM_UL(MEM(0xfa0000),0); */
#endif
	if (audio) audio_open();
	machine.screen_open();
#if MIDI
	if (midi) init_midi();
#endif
	init_hardware();
	init_ikbd();
#if MONITOR
	init_monitor();
#endif
	if (parallel_dev != NULL) init_parallel();
	if (serial_dev != NULL) init_serial();
#if 0
	load_file ("myrom.img", MEM(0x100000));
#endif
	if (verbose) fprintf(stderr,"Initializing CPU state...\n");
	init_cpu();
	fprintf(stderr,"Starting emulation...\n");
	execute_start(TOSSTART);
	kill_hardware();
#if MONITOR
	kill_monitor();
#endif
	machine.screen_close();
	if (parallel_dev != NULL) done_parallel();
	if (audio) audio_close();
	exit_debug();
	return 0;
}

